from typing import List

from rest_framework.request import Request

from bridger.buttons.bases import ButtonConfig as Button
from bridger.metadata.mixins import BridgerMetadataMixin


class CustomButtonMetadata(BridgerMetadataMixin):
    key = "custom_buttons"
    method_name = "_get_custom_buttons"


class CustomButtonMetadataMixin:
    def get_custom_buttons(self, request: Request, buttons: List) -> List:
        return buttons

    def _get_custom_buttons(self, request: Request) -> List:
        buttons = self.get_custom_buttons(
            request=request, buttons=getattr(self, "CUSTOM_BUTTONS", [])
        )
        button_list = list()
        for button in buttons:
            button.request = request
            button_list.append(dict(button))
        return button_list



from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Union

from bridger.display.formatting import Formatting, FormattingRule


@dataclass
class Field:
    key: str
    label: str
    formatting_rules: List[FormattingRule] = field(default_factory=list)
    width: Tuple[Union[float, str, int], str] = None

    def __iter__(self):
        yield "key", self.key
        yield "label", self.label
        yield "formatting_rules", [dict(rule) for rule in self.formatting_rules]

        if self.width:
            yield "width", self.width


@dataclass
class LegendItem:
    icon: str
    label: str
    value: Optional[str] = None

    def __iter__(self):
        yield "icon", self.icon
        yield "label", self.label

        if self.value:
            yield "value", self.value


@dataclass
class Legend:
    items: List[LegendItem]
    label: str = None
    key: str = None

    def __post_init__(self):
        if self.key:
            assert all(
                [bool(item.value) for item in self.items]
            ), "If key is set, all items need to specify a value."

    def __iter__(self):
        if self.label:
            yield "label", self.label

        if self.key:
            yield "key", self.key

        yield "items", [dict(item) for item in self.items]


@dataclass
class ListDisplay:
    fields: List[Field]
    legends: List[Legend] = field(default_factory=list)
    formatting: List[Formatting] = field(default_factory=list)

    def __iter__(self):
        yield "fields", [dict(field) for field in self.fields]
        yield "legends", [dict(legend) for legend in self.legends]
        yield "formatting", [dict(formatting) for formatting in self.formatting]


@dataclass
class Calendar:
    title: str
    start: str
    end: str
    filter_date_gte: str
    filter_date_lte: str

    def __iter__(self):
        yield "calendar", {
            "title": self.title,
            "start": self.start,
            "end": self.end,
            "filter_date_gte": self.filter_date_gte,
            "filter_date_lte": self.filter_date_lte,
        }



from calendar import monthrange
from datetime import date
from django.utils.timezone import localdate

from bridger.utils.date import get_start_date_from_date, get_end_date_from_date


def current_quarter_date_start(field, request):
     return get_start_date_from_date(localdate())


def current_quarter_date_end(field, request):
     return get_end_date_from_date(localdate())


def current_month_date_start(field, request):
     today = date.today()
     return date(today.year, today.month, 1)

def current_month_date_end(field, request):
     today = date.today()
     return date(today.year, today.month, monthrange(today.year, today.month)[1])


from rest_framework.request import Request

from bridger.enums import WidgetType
from bridger.metadata.mixins import BridgerMetadataMixin


class WidgetTypeMetadata(BridgerMetadataMixin):
    key = "type"
    method_name = "_get_widget_type"


class WidgetTypeMetadataMixin:
    def get_widget_type(self, request: Request, widget_type: str = None) -> str:
        return widget_type or (
            WidgetType.LIST.value
            if "pk" not in self.kwargs
            else WidgetType.INSTANCE.value
        )

    def _get_widget_type(self, request: Request) -> str:
        return self.get_widget_type(
            request=request, widget_type=getattr(self, "WIDGET_TYPE", None)
        )



from rest_framework.request import Request

from bridger.metadata.mixins import BridgerMetadataMixin


class PaginationMetadata(BridgerMetadataMixin):
    key = "pagination"
    method_name = "_get_pagination"


class PaginationMetadataMixin:
    def _get_pagination(self, request: Request):
        pagination = (
            self.pagination_class.__name__
            if hasattr(self, "pagination_class") and self.pagination_class
            else None
        )

        return {
            "CursorPagination": "cursor",
            "LimitOffsetPagination": "page",
            None: None,
        }[pagination]



from typing import Dict, Union

from rest_framework.request import Request

from bridger.display import Calendar, InstanceDisplay, ListDisplay
from bridger.metadata.mixins import BridgerMetadataMixin


class PKMetadata(BridgerMetadataMixin):
    key = "pk"
    method_name = "_get_pk"


class PKMetadataMixin:
    def get_pk(self, request: Request, pk=None):
        if pk:
            return pk

        return self.kwargs.get("pk", None)

    def _get_pk(self, request: Request):
        return self.get_pk(request=request, pk=getattr(self, "PK", None))



from dataclasses import dataclass
from enum import Enum


class MessageType(Enum):
    INFO = "info"
    WARNING = "warning"
    SUCCESS = "success"
    ERROR = "error"


@dataclass
class Message:
    message: str
    message_type: MessageType = MessageType.INFO
    auto_close: int = 5

    def __iter__(self):
        yield "message", self.message
        yield "type", self.message_type.value
        if self.auto_close:
            yield "auto_close", self.auto_close * 1000


def info(message, **kwargs):
    return Message(message=message, message_type=MessageType.INFO, **kwargs)


def warning(message, **kwargs):
    return Message(message=message, message_type=MessageType.WARNING, **kwargs)


def success(message, **kwargs):
    return Message(message=message, message_type=MessageType.SUCCESS, **kwargs)


def error(message, **kwargs):
    return Message(message=message, message_type=MessageType.ERROR, **kwargs)



from django.contrib import admin

from .models import FrontendUserConfiguration
from .notifications.admin import NotificationModelAdmin


class FrontendUserConfigurationInline(admin.TabularInline):
    model = FrontendUserConfiguration
    extra = 0
    fields = ["id", "user", "config"]
    readonly_fields = ["id", "user", "config"]


@admin.register(FrontendUserConfiguration)
class FrontendUserConfigurationModelAdmin(admin.ModelAdmin):
    list_display = ["id", "user"]
    fieldsets = (
        ("Main Information", {"fields": ("user", "parent_configuration", "config")}),
    )
    inlines = [FrontendUserConfigurationInline]



from dataclasses import dataclass
from typing import List, Tuple


@dataclass
class FieldSet:
    fields: List

    def __post_init__(self):
        assert isinstance(
            self.fields[0], (str, FieldSet)
        ), "fields can only contain strings or more FieldSets."

    def __iter__(self):
        for field in self.fields:
            if isinstance(field, str):
                yield field
            else:
                yield list(field)


@dataclass
class SectionList:
    key: str

    def __iter__(self):
        yield "key", self.key


@dataclass
class Section:
    fields: FieldSet = None
    section_list: SectionList = None
    title: str = None
    collapsed: bool = False

    def __post_init__(self):
        assert bool(self.fields) != bool(
            self.section_list
        ), "Either fields have to be provided or section_list, not both."

    def __iter__(self):
        if self.fields:
            yield "fields", [field for field in self.fields]
        elif self.section_list:
            yield "list", dict(self.section_list)

        yield "collapsed", self.collapsed

        if self.title:
            yield "title", self.title


@dataclass
class InstanceDisplay:
    sections: List[Section]

    def __iter__(self):
        yield from [dict(section) for section in self.sections]



from django.contrib import admin

from .models import Notification


@admin.register(Notification)
class NotificationModelAdmin(admin.ModelAdmin):
    readonly_fields = ["timestamp_created"]
    autocomplete_fields = ["recipient"]



from rest_framework import filters
from rest_framework.request import Request

from bridger.filters import DjangoFilterBackend
from bridger.metadata.mixins import BridgerMetadataMixin
from bridger.utils import ilen


class FilterFieldsMetadata(BridgerMetadataMixin):
    key = "filter_fields"
    method_name = "_get_filter_fields"


class FilterFieldsMetadataMixin:
    def get_filter_fields(self, request: Request):
        if DjangoFilterBackend in self.filter_backends:
            filterset_class = DjangoFilterBackend().get_filterset_class(
                self, self.queryset
            )
            if hasattr(filterset_class, "base_filters"):
                filterset = filterset_class.base_filters.items()
                field_names = [
                    v.field_name for k, v in filterset_class.base_filters.items()
                ]
                stored_fields = dict()
                for index, (name, field) in enumerate(filterset):

                    upcomming_fields = ilen(
                        filter(
                            lambda f: f == field.field_name, field_names[index + 1 :]
                        )
                    )
                    representation = field.get_representation(request, name)
                    
                    if upcomming_fields > 0 or field.field_name in stored_fields:
                        _representation = stored_fields.get(
                            field.field_name, representation
                        )
                        _representation["lookup_expr"].update(
                            representation["lookup_expr"]
                        )
                        _representation["default"].update(
                            representation["default"]
                        )
                        stored_fields[field.field_name] = _representation

                    if upcomming_fields == 0:
                        yield field.field_name, stored_fields.get(
                            field.field_name, representation
                        )

    def _get_filter_fields(self, request: Request):
        return dict(self.get_filter_fields(request))



from typing import Dict

from django.contrib.postgres.fields import ArrayField, JSONField
from django.db import models
from django_fsm import FSMField
from rest_framework import serializers
from rest_framework.request import Request

from bridger.fsm.mixins import FSMSerializerMetaclass
from bridger.serializers import fields

from .mixins import RepresentationSerializerMixin


def decorator(position: str, value: str) -> Dict:
    assert position in ("left", "right"), "Decorator Position can only be right or left"
    return {"position": position, "value": value}


percent_decorator = decorator(position="right", value="%")


class AdditionalMetadataMixin:
    @classmethod
    def get_decorators(cls):
        yield from getattr(cls.Meta, "decorators", dict()).items()

    @classmethod
    def get_percent_fields(cls):
        yield from getattr(cls.Meta, "percent_fields", list())


class Serializer(AdditionalMetadataMixin, serializers.Serializer):
    pass


class ModelSerializer(
    AdditionalMetadataMixin,
    serializers.ModelSerializer,
    metaclass=FSMSerializerMetaclass,
):

    serializer_field_mapping = {
        models.AutoField: fields.PrimaryKeyField,
        models.BooleanField: fields.BooleanField,
        models.CharField: fields.CharField,
        models.DateField: fields.DateField,
        models.DateTimeField: fields.DateTimeField,
        models.TimeField: fields.TimeField,
        models.DecimalField: fields.DecimalField,
        models.FileField: fields.FileField,
        models.FloatField: fields.FloatField,
        models.ImageField: fields.ImageField,
        models.IntegerField: fields.IntegerField,
        models.PositiveIntegerField: fields.IntegerField,
        models.PositiveSmallIntegerField: fields.IntegerField,
        models.SmallIntegerField: fields.IntegerField,
        models.TextField: fields.TextField,
        models.UUIDField: fields.CharField,
        ArrayField: fields.ListField,
        JSONField: fields.JSONField,
        fields.StarRatingField: fields.StarRatingField,
    }
    serializer_related_field = fields.PrimaryKeyRelatedField
    serializer_choice_field = fields.ChoiceField
    serializer_fsm_field = fields.FSMStatusField

    _additional_resources = fields.AdditionalResourcesField()
    _buttons = fields.DynamicButtonField()

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(
            field_name, model_field
        )
        if isinstance(model_field, FSMField):
            field_class = self.serializer_fsm_field

        return field_class, field_kwargs

    def build_property_field(self, field_name, model_class):
        field_class = fields.ReadOnlyField
        field_kwargs = {}

        return field_class, field_kwargs


class RepresentationSerializer(RepresentationSerializerMixin, ModelSerializer):
    def __init__(self, *args, **kwargs):
        self.filter_params = kwargs.pop(
            "filter_params", getattr(self, "filter_params", None)
        )

        self.endpoint = kwargs.pop(
            "endpoint",
            getattr(self, "endpoint", self.Meta.model.get_representation_endpoint()),
        )

        self.value_key = kwargs.pop(
            "value_key",
            getattr(self, "value_key", self.Meta.model.get_representation_value_key()),
        )

        self.label_key = kwargs.pop(
            "label_key",
            getattr(self, "label_key", self.Meta.model.get_representation_label_key()),
        )

        super().__init__(*args, **kwargs)



from .fields import (
    AdditionalResourcesField,
    DynamicButtonField,
    BooleanField,
    BridgerType,
    CharField,
    ChoiceField,
    DateField,
    DateTimeField,
    DecimalField,
    FileField,
    FloatField,
    FSMStatusField,
    HyperlinkField,
    ImageField,
    IntegerField,
    JSONTextEditorField,
    ListField,
    ListSerializer,
    PrimaryKeyCharField,
    PrimaryKeyField,
    PrimaryKeyRelatedField,
    RangeSelectField,
    ReadOnlyField,
    ReturnContentType,
    SerializerMethodField,
    StarRatingField,
    StringRelatedField,
    TextField,
    TimeField,
    register_resource,
    register_dynamic_button,
)
from .serializers import (
    Serializer,
    ModelSerializer,
    RepresentationSerializer,
    decorator,
    percent_decorator,
)



from channels.db import database_sync_to_async
from django.contrib.auth import get_user_model
from django.utils import timezone

from bridger.websockets.consumers import AsyncAuthenticatedJsonWebsocketConsumer

from .models import Notification


class NotificationConsumer(AsyncAuthenticatedJsonWebsocketConsumer):
    async def connect(self):
        await super().connect()
        if "user" in self.scope:
            channel_layer_name = f"notification-{self.scope['user'].id}"
            await self.channel_layer.group_add(channel_layer_name, self.channel_name)
            # await self.channel_layer.group_send(
            #     channel_layer_name,
            #     {"type": "notification.info", "user_id": self.scope["user"].id},
            # )

    async def get_notification_message(self, user):
        num_unreceived, num_unread = await database_sync_to_async(
            self.get_notification_information
        )(user)
        return f"You have {num_unreceived} new notifications and {num_unread} unread notifications."

    def get_notification_information(self, user):
        qs = Notification.objects.filter(recipient=user)
        num_unreceived = qs.filter(timestamp_received__isnull=True).update(
            timestamp_received=timezone.now()
        )
        num_unread = qs.filter(timestamp_read__isnull=True).count()

        return num_unreceived, num_unread

    async def notification_info(self, content):
        notification_message = await self.get_notification_message(
            get_user_model().objects.get(id=content["user_id"])
        )
        await self.send_json({"title": "Notification", "message": notification_message})

    async def notification_notify(self, content):
        await self.send_json(
            Notification.objects.get(id=content["notification_id"]).to_payload()
        )

    async def receive_json(self, content):
        """ Receives a dict in format of: {"notification_id": <int>} and marks it as received
        """
        database_sync_to_async(
            Notification.objects.filter(id=content["notification_id"]).update(
                timestamp_received=timezone.now()
            )
        )



from bridger.serializers import Serializer, CharField, IntegerField, TextField


class ShareSerializer(Serializer):

    user_id = IntegerField(label="User ID")
    widget_endpoint = CharField(label="Widget URL")
    message = TextField(label="Message", default="Check out this Widget.")




from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType, ReturnContentType


class CharField(BridgerSerializerFieldMixin, serializers.CharField):
    field_type = BridgerType.TEXT.value

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)


class StringRelatedField(BridgerSerializerFieldMixin, serializers.StringRelatedField):
    field_type = BridgerType.TEXT.value


class TextField(CharField):
    field_type = BridgerType.TEXTEDITOR.value
    texteditor_content_type = ReturnContentType.HTML.value

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["content_type"] = self.texteditor_content_type
        return representation



from .backends import DjangoFilterBackend
from .fields import (
    BooleanFilter,
    CharFilter,
    ChoiceFilter,
    DateFilter,
    DateTimeFilter,
    ModelChoiceFilter,
    ModelMultipleChoiceFilter,
    MultipleChoiceFilter,
    NumberFilter,
    TimeFilter,
)
from .filterset import FilterSet



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import close_old_connections
from jwt import decode as jwt_decode
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError
from rest_framework_simplejwt.tokens import UntypedToken


class JWTAuthMiddleware:
    """ Auth Middleware that authenticates a user with a JWT TOKEN """

    def __init__(self, inner):
        self.inner = inner

    def __call__(self, scope):
        # We should close all old connections here, somehow this bites itself with pytest
        # close_old_connections()

        try:
            jwt_access_token = scope["cookies"]["JWT-access"]
            UntypedToken(jwt_access_token)
        except (InvalidToken, TokenError, KeyError):
            return self.inner(dict(scope))
        else:
            decoded_data = jwt_decode(
                jwt_access_token, settings.SECRET_KEY, algorithms=["HS256"]
            )
            user = get_user_model().objects.get(id=decoded_data["user_id"])

        return self.inner(dict(scope, user=user))



from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType


class ChoiceField(BridgerSerializerFieldMixin, serializers.ChoiceField):
    field_type = BridgerType.SELECT.value

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["choices"] = [
            {"value": k, "label": v} for k, v in self.choices.items()
        ]
        return representation



class BridgerFilterMixin:
    def __init__(self, *args, **kwargs):
        self.default = kwargs.pop("default", None)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name):
        self.key = name
        representation = {
            "label": self.label,
            "type": self.filter_type,
            "key": self.field_name,
            "lookup_expr": {self.lookup_expr: self.key},
            "default": {},
        }

        if self.default:
            if callable(self.default):
                representation["default"][self.lookup_expr] = self.default(field=self, request=request)
            else:
                representation["default"][self.lookup_expr] = self.default

        return representation



from rest_framework.metadata import BaseMetadata

from bridger.enums import WidgetType
from bridger.history.viewsets import get_historical_viewset

from .fields import (
    ButtonMetadata,
    CreateButtonMetadata,
    CustomButtonMetadata,
    CustomInstanceButtonMetadata,
    EndpointMetadata,
    FieldsMetadata,
    FilterFieldsMetadata,
    IdentifierMetadata,
    InstanceDisplayMetadata,
    ListDisplayMetadata,
    OrderingFieldsMetadata,
    PaginationMetadata,
    PKMetadata,
    SearchFieldsMetadata,
    TitleMetadata,
    WidgetTypeMetadata,
    PreviewMetadata,
    DocsMetadata,
)


class BridgerMetadata(BaseMetadata):
    def __init__(self, *args, **kwargs):
        self.modules = [
            IdentifierMetadata,
            WidgetTypeMetadata,
            TitleMetadata,
            EndpointMetadata,
            ButtonMetadata,
            CreateButtonMetadata,
            CustomButtonMetadata,
            CustomInstanceButtonMetadata,
            PaginationMetadata,
            SearchFieldsMetadata,
            OrderingFieldsMetadata,
            FilterFieldsMetadata,
            PreviewMetadata,
            DocsMetadata,
        ]
        self.model_modules = [
            InstanceDisplayMetadata,
            ListDisplayMetadata,
            FieldsMetadata,
            PKMetadata,
        ]
        super().__init__(*args, **kwargs)

    def generate_metadata(self, request, view):
        for module in self.modules:
            yield from module(view, request)

    def generate_model_metadata(self, request, view):
        for module in self.model_modules:
            yield from module(view, request)

    def determine_metadata(self, request, view):
        if view.historical_mode:
            format_kwarg = view.format_kwarg
            old_kwargs = view.kwargs

            model = type(view.get_object())
            historical_model = view.get_object().history.model

            view_class = get_historical_viewset(model, historical_model)
            view = view_class()
            view.kwargs = {"model_pk": old_kwargs["pk"]}
            view.request = request
            view.format_kwarg = format_kwarg

        metadata = dict(self.generate_metadata(request, view))

        if metadata["type"] in [WidgetType.INSTANCE.value, WidgetType.LIST.value]:
            metadata.update(dict(self.generate_model_metadata(request, view)))

        return metadata



from rest_framework.request import Request

from bridger.enums import WidgetType
from bridger.metadata.mixins import BridgerMetadataMixin


class TitleMetadata(BridgerMetadataMixin):
    key = "titles"
    method_name = "_get_titles"


class TitleMetadataMixin:
    def get_instance_title(self, request: Request, title: str = None) -> str:
        if title:
            return title

        if hasattr(self, "get_serializer_class"):
            model = self.get_serializer_class().Meta.model

            if hasattr(model, "get_representation_label_key"):
                return f"{model._meta.verbose_name}: {model.get_representation_label_key()}"

            return model._meta.verbose_name

        return ""

    def get_list_title(self, request: Request, title: str = None) -> str:
        if title:
            return title

        if hasattr(self, "get_serializer_class"):
            model = self.get_serializer_class().Meta.model
            return model._meta.verbose_name_plural

        return ""

    def get_create_title(self, request: Request, title: str = None) -> str:
        if title:
            return title

        if hasattr(self, "get_serializer_class"):
            model = self.get_serializer_class().Meta.model
            return f"Create {model._meta.verbose_name}"

        return ""

    def _get_titles(self, request: Request) -> str:
        return {
            "instance": self.get_instance_title(
                request, getattr(self, "INSTANCE_TITLE", None)
            ),
            "list": self.get_list_title(request, getattr(self, "LIST_TITLE", None)),
            "create": self.get_create_title(
                request, getattr(self, "CREATE_TITLE", None)
            ),
        }



import logging

import django_filters
from django.conf import settings
from django.db.models import QuerySet
from django.urls.exceptions import NoReverseMatch
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, status, viewsets
from rest_framework.decorators import (
    action,
    api_view,
    permission_classes,
    renderer_classes,
)
from rest_framework.mixins import ListModelMixin
from rest_framework.permissions import IsAuthenticated
from rest_framework.renderers import StaticHTMLRenderer
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse

from .docs import get_markdown_docs
from .enums import WidgetType
from .filters import BooleanFilter, ModelChoiceFilter
from .fsm.mixins import FSMViewSetMixin

# from .mixins import MetadataMixin
from .metadata.views import MetadataMixin
from .pagination import CursorPagination
from .settings import bridger_settings
from .history.serializers import get_historical_serializer

logger = logging.getLogger(__name__)
from django.shortcuts import get_object_or_404


class InstanceMixin:
    def retrieve(self, request, *args, **kwargs):

        instance = self.get_object()
        serializer = self.get_serializer(instance)
        serialized_content = {"instance": serializer.data}

        if hasattr(self, "get_messages"):
            messages = self.get_messages(request=request, instance=instance)
            if messages:
                serialized_content["messages"] = [dict(message) for message in messages]

        return Response(serialized_content)

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return Response({"instance": serializer.data})


class RepresentationModelViewSet(
    MetadataMixin, InstanceMixin, viewsets.ReadOnlyModelViewSet
):
    """A Representation View that is used for serializing related fields"""

    filter_backends = [
        filters.OrderingFilter,
        filters.SearchFilter,
        DjangoFilterBackend,
    ]
    pagination_class = CursorPagination

    filter_fields = {}
    search_fields = []
    ordering_fields = ordering = ["id"]
    historical_mode = False


class ModelViewSet(
    MetadataMixin, InstanceMixin, FSMViewSetMixin, viewsets.ModelViewSet
):
    """A Model View that is used to serializer models"""

    filter_backends = [
        filters.OrderingFilter,
        filters.SearchFilter,
        DjangoFilterBackend,
    ]
    pagination_class = CursorPagination

    filter_fields = {}
    search_fields = []
    ordering_fields = ordering = ["id"]
    historical_mode = False

    @classmethod
    def get_model(cls):
        try:
            if hasattr(cls, "queryset"):
                return cls.queryset.model
            elif hasattr(cls, "serializer_class"):
                return cls.serializer_class.Meta.model
            else:
                return None
        except AttributeError:
            return None

    def destroy_multiple(self, request, *args, **kwargs):
        model = self.get_serializer_class().Meta.model
        app_label = model._meta.app_label

        queryset = model.objects.filter(id__in=request.data)
        destroyed = self.perform_destroy_multiple(queryset)

        return Response(
            {"count": destroyed[1].get(f"{app_label}.{model.__name__}", 0)},
            status=status.HTTP_204_NO_CONTENT,
        )

    def perform_destroy_multiple(self, queryset):
        return queryset.delete()

    def get_messages(
        self,
        request,
        queryset=None,
        paginated_queryset=None,
        instance=None,
        initial=False,
    ):
        return []

    def get_serializer_changes(self, serializer):
        return serializer

    def get_serializer(self, *args, **kwargs):
        return self.get_serializer_changes(super().get_serializer(*args, **kwargs))

    @action(
        methods=["get"],
        detail=False,
        url_name="list-docs",
        renderer_classes=[StaticHTMLRenderer],
    )
    def __list_docs__(self, request, *args, **kwargs):
        return get_markdown_docs(self.LIST_DOCS)

    @action(
        methods=["get"],
        detail=True,
        url_name="instance-docs",
        renderer_classes=[StaticHTMLRenderer],
    )
    def __instance_docs__(self, request, *args, **kwargs):
        return get_markdown_docs(self.INSTANCE_DOCS)

    def history_list(self, request, *args, **kwargs):
        obj = self.get_object()
        serializer_class = get_historical_serializer(obj.history.model)
        serializer = serializer_class(obj.history.all(), many=True)

        return Response({"results": serializer.data})

    def history_retrieve(self, request, *args, **kwargs):
        obj = self.get_object()
        model = type(obj)
        opts = model._meta

        historical_model = getattr(model, opts.simple_history_manager_attribute).model
        historical_opts = historical_model._meta
        historical_obj = get_object_or_404(
            historical_model,
            **{opts.pk.attname: obj.id, "history_id": kwargs["history_id"]},
        )

        for field in filter(
            lambda f: not f.attname.startswith("history_"), obj._meta.fields
        ):
            setattr(obj, field.attname, getattr(historical_obj, field.attname))

        serializer = self.get_serializer(obj)
        return Response({"instance": serializer.data})


class InfiniteDataModelView(ModelViewSet):

    pagination_class = None
    historical_mode = False

    def list(self, request, *args, **kwargs) -> Response:
        queryset = self.filter_queryset(self.get_queryset())
        serializer = self.get_serializer(queryset, many=True)

        aggregates = dict()
        messages = dict()
        if hasattr(self, "get_aggregates"):
            aggregates = self.get_aggregates(queryset, queryset)

        if hasattr(self, "get_messages"):
            messages = self.get_messages(
                request=self.request, queryset=queryset, paginated_queryset=queryset
            )

        return Response(
            {"results": serializer.data, "aggregates": aggregates, "messages": messages}
        )


class ChartViewSet(MetadataMixin, ListModelMixin, viewsets.ViewSet):
    """A List View that is used for creating plotly charts"""

    WIDGET_TYPE = WidgetType.CHART.value
    filter_backends = [
        filters.OrderingFilter,
        filters.SearchFilter,
        DjangoFilterBackend,
    ]
    pagination_class = CursorPagination

    filter_fields = {}
    search_fields = []
    ordering_fields = []
    historical_mode = False

    def list(self, request, *args, **kwargs):
        figure = self.get_plotly(self.filter_queryset(self.get_queryset()))
        figure_dict = figure.to_plotly_json()
        figure_dict["config"] = {"responsive": True, "displaylogo": False}
        figure_dict["useResizeHandler"] = True
        figure_dict["style"] = {"width": "100%", "height": "100%"}

        return Response(figure_dict)

    def get_queryset(self):
        return self.queryset

    def filter_queryset(self, queryset: QuerySet) -> QuerySet:
        for backend in list(self.filter_backends):
            queryset = backend().filter_queryset(self.request, queryset, self)
        return queryset


@api_view(["GET"])
@permission_classes([IsAuthenticated])
def api_endpoints_root(request: Request, format: str = None) -> Response:
    """Returns a list of API-ROOT endpoints specified in BRIDGER_ENDPOINTS"""

    endpoints = dict()
    endpoints["config"] = reverse("bridger:config", request=request, format=format)

    if hasattr(settings, "BRIDGER_ENDPOINTS"):
        for bridger_endpoint in settings.BRIDGER_ENDPOINTS:
            try:
                endpoints[bridger_endpoint] = reverse(
                    f"{bridger_endpoint}:api-root", request=request, format=format
                )
            except NoReverseMatch:
                endpoints[bridger_endpoint] = reverse(
                    f"API-ROOT of {bridger_endpoint} is not yet implemented."
                )

    return Response(endpoints, status=status.HTTP_200_OK)
